from __future__ import annotations

import json
import logging
import os
import time
from datetime import datetime, timedelta, timezone
from collections import Counter
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple

from sqlalchemy import delete, func, select, or_, update, insert, text
from sqlalchemy.exc import DataError, ProgrammingError, IntegrityError
from sqlalchemy.ext.asyncio import AsyncSession

from fmatch.saas.model_defs.salesforce_schema_index import (
    SalesforceSchemaField,
    SalesforceSchemaFieldAppUsage,
    SalesforceSchemaFieldProfile,
    SalesforceSchemaFieldUsage,
    SalesforceSchemaIndexState,
    SalesforceSchemaObject,
    SalesforceSchemaRelationship,
    SalesforceSchemaTerm,
)

logger = logging.getLogger(__name__)
EMBEDDING_RECHECK_SECONDS = int(os.getenv("SF_EMBEDDING_RECHECK_S", "300"))


class SalesforceSchemaIndexRepo:
    """Persistence helpers for the Salesforce schema index."""

    def __init__(self, db: AsyncSession) -> None:
        self.db = db
        self._embedding_column_available: Optional[bool] = None
        self._embedding_column_retry_at: Optional[float] = None
        self._embedding_recheck_seconds = EMBEDDING_RECHECK_SECONDS

    @staticmethod
    def _utcnow_naive() -> datetime:
        """Return a UTC datetime without tzinfo to match TIMESTAMP WITHOUT TIME ZONE columns."""
        return datetime.now(timezone.utc).replace(tzinfo=None)

    def _embedding_available(self) -> bool:
        if self._embedding_column_available is False:
            if (
                self._embedding_column_retry_at is not None
                and time.monotonic() >= self._embedding_column_retry_at
            ):
                self._embedding_column_available = None
                self._embedding_column_retry_at = None
                logger.info("Rechecking embedding column after cooldown")
            else:
                return False
        return True

    @staticmethod
    def _as_naive(dt: Optional[datetime]) -> Optional[datetime]:
        if dt is None:
            return None
        if dt.tzinfo:
            return dt.replace(tzinfo=None)
        return dt

    # ------------------------------------------------------------------ #
    # Index state helpers
    # ------------------------------------------------------------------ #
    async def get_state(self, org_id: str) -> Optional[SalesforceSchemaIndexState]:
        return await self.db.get(SalesforceSchemaIndexState, org_id)

    async def get_status_row(self, org_id: str) -> Optional[SalesforceSchemaIndexState]:
        return await self.get_state(org_id)

    async def mark_in_progress(
        self, org_id: str, *, stage: str, token: str
    ) -> SalesforceSchemaIndexState:
        now = self._utcnow_naive()
        stale_cutoff = now - timedelta(minutes=5)

        update_stmt = (
            update(SalesforceSchemaIndexState)
            .where(SalesforceSchemaIndexState.org_id == org_id)
            .where(
                or_(
                    SalesforceSchemaIndexState.in_progress == False,
                    SalesforceSchemaIndexState.heartbeat_at.is_(None),
                    SalesforceSchemaIndexState.heartbeat_at < stale_cutoff,
                )
            )
            .values(
                status=f"{stage}_running",
                in_progress=True,
                stage=stage,
                runner_token=token,
                heartbeat_at=now,
                updated_at=now,
            )
        )
        update_result = await self.db.execute(update_stmt)
        if (update_result.rowcount or 0) == 1:
            row = await self.get_state(org_id)
            if row:
                return row

        row = SalesforceSchemaIndexState(
            org_id=org_id,
            schema_hash=None,
            status=f"{stage}_running",
            in_progress=True,
            stage=stage,
            runner_token=token,
            heartbeat_at=now,
            checkpoint_json="{}",
            updated_at=now,
        )
        self.db.add(row)
        try:
            await self.db.flush()
            return row
        except IntegrityError as exc:
            await self.db.rollback()
            raise RuntimeError("Index already running") from exc

    async def clear_in_progress(self, org_id: str, *, token: Optional[str]) -> None:
        row = await self.get_state(org_id)
        if not row:
            return
        if token and row.runner_token and row.runner_token != token:
            return
        row.in_progress = False
        row.stage = None
        row.runner_token = None
        row.heartbeat_at = None
        if row.status.endswith("_running"):
            row.status = "idle"
        await self.db.flush()

    async def touch_heartbeat(self, org_id: str, *, token: Optional[str]) -> None:
        # Token is required to prevent heartbeat hijacking
        if not token:
            raise ValueError("Token required for heartbeat - prevents unauthorized lock takeover")
        stmt = (
            update(SalesforceSchemaIndexState)
            .where(
                SalesforceSchemaIndexState.org_id == org_id,
                SalesforceSchemaIndexState.runner_token == token,
            )
            .values(heartbeat_at=self._utcnow_naive())
        )
        await self.db.execute(stmt)

    async def get_checkpoint(self, org_id: str) -> Dict[str, Any]:
        row = await self.get_state(org_id)
        if not row or not row.checkpoint_json:
            return {}
        try:
            return json.loads(row.checkpoint_json)
        except Exception:  # noqa: BLE001
            return {}

    async def set_checkpoint(
        self, org_id: str, checkpoint: Optional[Dict[str, Any]]
    ) -> None:
        row = await self.get_state(org_id)
        payload = json.dumps(checkpoint or {}, ensure_ascii=False)
        if not row:
            row = SalesforceSchemaIndexState(
                org_id=org_id,
                checkpoint_json=payload,
            )
            self.db.add(row)
        else:
            row.checkpoint_json = payload
        await self.db.flush()

    async def mark_quick_indexed(
        self,
        org_id: str,
        *,
        schema_hash: Optional[str],
        fls_hash: Optional[str] = None,
    ) -> None:
        now = self._utcnow_naive()
        row = await self.get_state(org_id)
        if not row:
            row = SalesforceSchemaIndexState(
                org_id=org_id,
                schema_hash=schema_hash,
                quick_indexed_at=now,
                status="ready",
                fls_hash=fls_hash,
            )
            self.db.add(row)
        else:
            row.schema_hash = schema_hash
            row.quick_indexed_at = now
            row.status = "ready"
            if fls_hash:
                row.fls_hash = fls_hash
        row.in_progress = False
        row.stage = None
        row.runner_token = None
        row.heartbeat_at = None
        await self.db.flush()

    async def upsert_state(
        self,
        org_id: str,
        *,
        schema_hash: Optional[str],
        status: str,
        quick_indexed_at: Optional[datetime] = None,
        full_indexed_at: Optional[datetime] = None,
        error_json: Optional[str] = None,
    ) -> SalesforceSchemaIndexState:
        quick_indexed_at = self._as_naive(quick_indexed_at)
        full_indexed_at = self._as_naive(full_indexed_at)
        row = await self.db.get(SalesforceSchemaIndexState, org_id)
        if not row:
            row = SalesforceSchemaIndexState(
                org_id=org_id,
                schema_hash=schema_hash,
                status=status,
                quick_indexed_at=quick_indexed_at,
                full_indexed_at=full_indexed_at,
                error_json=error_json,
            )
            self.db.add(row)
        else:
            row.schema_hash = schema_hash
            row.status = status
            if quick_indexed_at is not None:
                row.quick_indexed_at = quick_indexed_at
            if full_indexed_at is not None:
                row.full_indexed_at = full_indexed_at
            row.error_json = error_json
            row.updated_at = self._utcnow_naive()
            if not row.checkpoint_json:
                row.checkpoint_json = "{}"
        await self.db.flush()
        return row

    # ------------------------------------------------------------------ #
    # Bulk persistence helpers
    # ------------------------------------------------------------------ #
    async def replace_objects(
        self, org_id: str, schema_hash: str, rows: Iterable[SalesforceSchemaObject]
    ) -> None:
        rows = list(rows)
        if not self._embedding_available():
            await self._replace_objects_without_embedding(org_id, rows)
            return
        try:
            async with self.db.begin_nested():
                await self.db.execute(
                    delete(SalesforceSchemaObject).where(
                        SalesforceSchemaObject.org_id == org_id
                    )
                )
                self.db.add_all(rows)
                await self.db.flush()
        except ProgrammingError as exc:
            # Savepoint auto-rollback handled by begin_nested() context manager
            if not self._handle_missing_embedding(exc, stage="objects"):
                raise
            await self._replace_objects_without_embedding(org_id, rows)
        except DataError as exc:
            # Savepoint auto-rollback handled by begin_nested() context manager
            logger.error("Data error in replace_objects for %s: %s", org_id, exc)
            raise

    async def replace_fields(
        self,
        org_id: str,
        schema_hash: str,
        rows: Iterable[SalesforceSchemaField],
    ) -> None:
        rows = list(rows)
        if not self._embedding_available():
            await self._replace_fields_without_embedding(org_id, rows)
            return
        try:
            async with self.db.begin_nested():
                await self.db.execute(
                    delete(SalesforceSchemaField).where(
                        SalesforceSchemaField.org_id == org_id
                    )
                )
                self.db.add_all(rows)
                await self.db.flush()
        except ProgrammingError as exc:
            # Savepoint auto-rollback handled by begin_nested() context manager
            if not self._handle_missing_embedding(exc, stage="fields"):
                raise
            await self._replace_fields_without_embedding(org_id, rows)
        except DataError as exc:
            # Savepoint auto-rollback handled by begin_nested() context manager
            logger.error("Data error in replace_fields for %s: %s", org_id, exc)
            raise

    async def replace_relationships(
        self,
        org_id: str,
        schema_hash: str,
        rows: Iterable[SalesforceSchemaRelationship],
    ) -> None:
        try:
            async with self.db.begin_nested():
                await self.db.execute(
                    delete(SalesforceSchemaRelationship).where(
                        SalesforceSchemaRelationship.org_id == org_id
                    )
                )
                self.db.add_all(list(rows))
                await self.db.flush()
        except DataError as exc:
            # Savepoint auto-rollback handled by begin_nested() context manager
            logger.error("Data error in replace_relationships for %s: %s", org_id, exc)
            raise

    async def replace_terms(
        self,
        org_id: str,
        schema_hash: str,
        rows: Iterable[SalesforceSchemaTerm],
    ) -> None:
        try:
            async with self.db.begin_nested():
                await self.db.execute(
                    delete(SalesforceSchemaTerm).where(SalesforceSchemaTerm.org_id == org_id)
                )
                self.db.add_all(list(rows))
                await self.db.flush()
        except DataError as exc:
            # Savepoint auto-rollback handled by begin_nested() context manager
            logger.error("Data error in replace_terms for %s: %s", org_id, exc)
            raise

    # Usage / profile upserters
    async def upsert_field_usage(
        self,
        *,
        org_id: str,
        field_ref: str,
        report_count: int,
        last_run_at: Optional[datetime],
        last_referenced_at: Optional[datetime],
        schema_hash: Optional[str],
    ) -> None:
        stmt = select(SalesforceSchemaFieldUsage).where(
            SalesforceSchemaFieldUsage.org_id == org_id,
            SalesforceSchemaFieldUsage.field_ref == field_ref,
        )
        existing = (await self.db.execute(stmt)).scalar_one_or_none()
        if existing:
            existing.report_count = report_count
            existing.last_run_at = last_run_at
            existing.last_referenced_at = last_referenced_at
            existing.schema_hash = schema_hash
            existing.indexed_at = self._utcnow_naive()
        else:
            row = SalesforceSchemaFieldUsage(
                org_id=org_id,
                field_ref=field_ref,
                report_count=report_count,
                last_run_at=last_run_at,
                last_referenced_at=last_referenced_at,
                schema_hash=schema_hash,
            )
            self.db.add(row)
        await self.db.flush()

    async def upsert_field_profile(
        self,
        *,
        org_id: str,
        field_ref: str,
        not_null_rate: Optional[float],
        distinct_ratio: Optional[float],
        top_values: Optional[Sequence[str]] = None,
        sample_values: Optional[Sequence[str]] = None,
    ) -> None:
        stmt = select(SalesforceSchemaFieldProfile).where(
            SalesforceSchemaFieldProfile.org_id == org_id,
            SalesforceSchemaFieldProfile.field_ref == field_ref,
        )
        existing = (await self.db.execute(stmt)).scalar_one_or_none()
        payload = json.dumps(list(top_values or []), ensure_ascii=False)
        sample = json.dumps(list(sample_values or []), ensure_ascii=False)
        if existing:
            existing.not_null_rate = not_null_rate
            existing.distinct_ratio = distinct_ratio
            existing.top_values_json = payload
            existing.sample_values_json = sample
            existing.updated_at = self._utcnow_naive()
        else:
            row = SalesforceSchemaFieldProfile(
                org_id=org_id,
                field_ref=field_ref,
                not_null_rate=not_null_rate,
                distinct_ratio=distinct_ratio,
                top_values_json=payload,
                sample_values_json=sample,
            )
            self.db.add(row)
        await self.db.flush()

    async def upsert_app_usage(
        self,
        *,
        org_id: str,
        field_ref: str,
        query_delta: int,
        preview_delta: int,
        cooccur: Optional[dict] = None,
    ) -> None:
        stmt = select(SalesforceSchemaFieldAppUsage).where(
            SalesforceSchemaFieldAppUsage.org_id == org_id,
            SalesforceSchemaFieldAppUsage.field_ref == field_ref,
        )
        existing = (await self.db.execute(stmt)).scalar_one_or_none()
        if existing:
            existing.query_count_30d = max(0, existing.query_count_30d + query_delta)
            existing.preview_count_30d = max(
                0, existing.preview_count_30d + preview_delta
            )
            if cooccur:
                current = json.loads(existing.cooccur_json or "{}")
                for k, v in cooccur.items():
                    current[k] = current.get(k, 0) + v
                existing.cooccur_json = json.dumps(current, ensure_ascii=False)
            existing.updated_at = self._utcnow_naive()
        else:
            row = SalesforceSchemaFieldAppUsage(
                org_id=org_id,
                field_ref=field_ref,
                query_count_30d=max(0, query_delta),
                preview_count_30d=max(0, preview_delta),
                cooccur_json=json.dumps(cooccur or {}, ensure_ascii=False),
            )
            self.db.add(row)
        await self.db.flush()

    # ------------------------------------------------------------------ #
    # Query helpers for discover/planner
    # ------------------------------------------------------------------ #
    async def list_objects(
        self, org_id: str, limit: int = 50
    ) -> List[SalesforceSchemaObject]:
        stmt = (
            select(SalesforceSchemaObject)
            .where(SalesforceSchemaObject.org_id == org_id)
            .order_by(
                SalesforceSchemaObject.is_custom.asc(),
                SalesforceSchemaObject.label.asc(),
            )
            .limit(limit)
        )
        return list((await self.db.execute(stmt)).scalars().all())

    async def list_fields(
        self, org_id: str, object_api: str
    ) -> List[SalesforceSchemaField]:
        stmt = (
            select(SalesforceSchemaField)
            .where(
                SalesforceSchemaField.org_id == org_id,
                SalesforceSchemaField.object_api == object_api,
            )
            .order_by(SalesforceSchemaField.field_api.asc())
        )
        return list((await self.db.execute(stmt)).scalars().all())

    async def top_terms(
        self, org_id: str, term: str, limit: int = 10
    ) -> List[Tuple[str, float]]:
        stmt = (
            select(SalesforceSchemaTerm.field_ref, SalesforceSchemaTerm.weight)
            .where(
                SalesforceSchemaTerm.org_id == org_id,
                SalesforceSchemaTerm.term == term.lower(),
            )
            .order_by(SalesforceSchemaTerm.weight.desc())
            .limit(limit)
        )
        return list((await self.db.execute(stmt)).all())

    async def replace_field_usage(self, org_id: str, usage: Counter) -> None:
        try:
            async with self.db.begin_nested():
                await self.db.execute(
                    delete(SalesforceSchemaFieldUsage).where(
                        SalesforceSchemaFieldUsage.org_id == org_id
                    )
                )
                now = self._utcnow_naive()
                rows = [
                    SalesforceSchemaFieldUsage(
                        org_id=org_id,
                        field_ref=field_ref,
                        report_count=int(count),
                        indexed_at=now,
                    )
                    for field_ref, count in usage.items()
                ]
                if rows:
                    self.db.add_all(rows)
                await self.db.flush()
        except DataError as exc:
            # Savepoint auto-rollback handled by begin_nested() context manager
            logger.error("Data error in replace_field_usage for %s: %s", org_id, exc)
            raise

    async def save_report_usage(
        self, org_id: str, usage: Counter, cooccur: Dict[str, Counter]
    ) -> None:
        await self.replace_field_usage(org_id, usage)
        # co-occurrence matrix is kept in-memory for scoring; persist in future if required

    async def replace_field_profiles(
        self, org_id: str, profiles: Dict[str, Dict[str, Dict[str, float]]]
    ) -> None:
        now = self._utcnow_naive()
        rows = []
        for obj, field_map in profiles.items():
            for field_api, stats in field_map.items():
                field_ref = f"{obj}.{field_api}"
                rows.append(
                    SalesforceSchemaFieldProfile(
                        org_id=org_id,
                        field_ref=field_ref,
                        not_null_rate=stats.get("not_null_rate"),
                        distinct_ratio=stats.get("distinct_ratio"),
                        top_values_json=json.dumps(
                            stats.get("top_values", []) or [], ensure_ascii=False
                        ),
                        sample_values_json=json.dumps(
                            stats.get("sample_values", []) or [], ensure_ascii=False
                        ),
                        updated_at=now,
                    )
                )
        try:
            async with self.db.begin_nested():
                await self.db.execute(
                    delete(SalesforceSchemaFieldProfile).where(
                        SalesforceSchemaFieldProfile.org_id == org_id
                    )
                )
                if rows:
                    self.db.add_all(rows)
                await self.db.flush()
        except DataError as exc:
            # Savepoint auto-rollback handled by begin_nested() context manager
            logger.error("Data error in replace_field_profiles for %s: %s", org_id, exc)
            raise

    async def update_term_weights(self, org_id: str, scores: Dict[str, float]) -> None:
        if not scores:
            return
        stmt = select(SalesforceSchemaTerm).where(SalesforceSchemaTerm.org_id == org_id)
        rows = (await self.db.execute(stmt)).scalars().all()
        now = self._utcnow_naive()
        for row in rows:
            score = scores.get(row.field_ref)
            if score is None:
                continue
            row.weight = float(max(score, row.weight or 0.0))
            row.indexed_at = now
        await self.db.flush()

    async def app_usage(self, org_id: str) -> Dict[str, float]:
        stmt = select(
            SalesforceSchemaFieldAppUsage.field_ref,
            SalesforceSchemaFieldAppUsage.query_count_30d,
            SalesforceSchemaFieldAppUsage.preview_count_30d,
        ).where(SalesforceSchemaFieldAppUsage.org_id == org_id)
        return {
            field_ref: float((query or 0) + (preview or 0))
            for field_ref, query, preview in await self.db.execute(stmt)
        }

    async def cooccur_sum(self, org_id: str, field_ref: str) -> float:
        stmt = select(SalesforceSchemaFieldAppUsage.cooccur_json).where(
            SalesforceSchemaFieldAppUsage.org_id == org_id,
            SalesforceSchemaFieldAppUsage.field_ref == field_ref,
        )
        raw = (await self.db.execute(stmt)).scalar_one_or_none()
        if not raw:
            return 0.0
        try:
            data = json.loads(raw)
        except Exception:
            return 0.0
        if isinstance(data, dict):
            return float(sum(data.values()))
        return 0.0

    async def all_field_refs(self, org_id: str) -> List[Tuple[str, str, bool]]:
        stmt = select(
            SalesforceSchemaField.object_api,
            SalesforceSchemaField.field_api,
            SalesforceSchemaField.is_custom,
        ).where(SalesforceSchemaField.org_id == org_id)
        return [(row[0], row[1], bool(row[2])) for row in await self.db.execute(stmt)]

    async def queryable_fields(
        self, org_id: str, object_api: str, limit: int = 120
    ) -> List[str]:
        stmt = (
            select(SalesforceSchemaField.field_api)
            .where(
                SalesforceSchemaField.org_id == org_id,
                SalesforceSchemaField.object_api == object_api,
                SalesforceSchemaField.queryable.is_(True),
            )
            .limit(limit)
        )
        return [row[0] for row in await self.db.execute(stmt)]

    # ------------------------------------------------------------------ #
    # Semantic search helpers (PGVector)
    # ------------------------------------------------------------------ #
    async def find_similar_objects(
        self, org_id: str, vector: List[float], limit: int = 5
    ) -> List[SalesforceSchemaObject]:
        if not vector:
            return []
        if not self._embedding_available():
            return []
        # Use raw SQL since embedding column is not in ORM model
        vector_str = "[" + ",".join(str(v) for v in vector) + "]"
        sql = text("""
            SELECT org_id, object_api FROM sf_schema_objects
            WHERE org_id = :org_id AND embedding IS NOT NULL
            ORDER BY embedding <=> CAST(:vector AS vector)
            LIMIT :limit
        """)
        try:
            result = await self.db.execute(
                sql, {"org_id": org_id, "vector": vector_str, "limit": limit}
            )
            object_apis = [row[1] for row in result.fetchall()]
            if not object_apis:
                return []
            # Fetch full objects by their APIs
            stmt = select(SalesforceSchemaObject).where(
                SalesforceSchemaObject.org_id == org_id,
                SalesforceSchemaObject.object_api.in_(object_apis),
            )
            return list((await self.db.execute(stmt)).scalars().all())
        except ProgrammingError as exc:
            if self._handle_missing_embedding(exc, stage="objects"):
                return []
            raise

    async def find_similar_fields(
        self,
        org_id: str,
        vector: List[float],
        limit: int = 20,
        object_filter: Optional[List[str]] = None,
    ) -> List[SalesforceSchemaField]:
        if not vector:
            return []
        if not self._embedding_available():
            return []
        # Use raw SQL since embedding column is not in ORM model
        vector_str = "[" + ",".join(str(v) for v in vector) + "]"
        if object_filter:
            sql = text("""
                SELECT org_id, object_api, field_api FROM sf_schema_fields
                WHERE org_id = :org_id AND embedding IS NOT NULL
                  AND object_api = ANY(:object_filter)
                ORDER BY embedding <=> CAST(:vector AS vector)
                LIMIT :limit
            """)
            params = {
                "org_id": org_id,
                "vector": vector_str,
                "limit": limit,
                "object_filter": object_filter,
            }
        else:
            sql = text("""
                SELECT org_id, object_api, field_api FROM sf_schema_fields
                WHERE org_id = :org_id AND embedding IS NOT NULL
                ORDER BY embedding <=> CAST(:vector AS vector)
                LIMIT :limit
            """)
            params = {"org_id": org_id, "vector": vector_str, "limit": limit}
        try:
            result = await self.db.execute(sql, params)
            rows = result.fetchall()
            if not rows:
                return []
            # Fetch full fields by their keys
            conditions = [
                (SalesforceSchemaField.org_id == org_id)
                & (SalesforceSchemaField.object_api == r[1])
                & (SalesforceSchemaField.field_api == r[2])
                for r in rows
            ]
            stmt = select(SalesforceSchemaField).where(or_(*conditions))
            return list((await self.db.execute(stmt)).scalars().all())
        except ProgrammingError as exc:
            if self._handle_missing_embedding(exc, stage="fields"):
                return []
            raise

    def _handle_missing_embedding(self, exc: Exception, *, stage: str) -> bool:
        message = str(exc).lower()
        if "embedding" in message and "does not exist" in message:
            if self._embedding_column_available is not False:
                self._embedding_column_available = False
                if self._embedding_recheck_seconds > 0:
                    self._embedding_column_retry_at = (
                        time.monotonic() + self._embedding_recheck_seconds
                    )
                logger.warning(
                    "Schema index embeddings disabled (%s): missing embedding column. "
                    "Run migrations to add pgvector columns.",
                    stage,
                )
            return True
        return False

    async def _replace_objects_without_embedding(
        self, org_id: str, rows: List[SalesforceSchemaObject]
    ) -> None:
        async with self.db.begin_nested():
            await self.db.execute(
                delete(SalesforceSchemaObject).where(
                    SalesforceSchemaObject.org_id == org_id
                )
            )
            # embedding column exists in DB but not in ORM model
            payload = [row.model_dump() for row in rows]
            if payload:
                await self.db.execute(
                    insert(SalesforceSchemaObject.__table__).values(payload)
                )
            await self.db.flush()

    async def _replace_fields_without_embedding(
        self, org_id: str, rows: List[SalesforceSchemaField]
    ) -> None:
        async with self.db.begin_nested():
            await self.db.execute(
                delete(SalesforceSchemaField).where(
                    SalesforceSchemaField.org_id == org_id
                )
            )
            # embedding column exists in DB but not in ORM model
            payload = [row.model_dump() for row in rows]
            if payload:
                await self.db.execute(
                    insert(SalesforceSchemaField.__table__).values(payload)
                )
            await self.db.flush()

    async def rank_for_terms(
        self,
        *,
        org_id: str,
        terms: Sequence[str],
        limit: int = 20,
    ) -> List[Tuple[str, float, str, str, str, int, float]]:
        normalized_terms = [term.lower() for term in terms if term]
        seen: Dict[str, float] = {}
        results: List[Tuple[str, float, str, str, str, int, float]] = []

        for term in normalized_terms or ["amount"]:
            term_rows = (
                await self.db.execute(
                    select(
                        SalesforceSchemaTerm.field_ref,
                        SalesforceSchemaTerm.weight,
                    )
                    .where(
                        SalesforceSchemaTerm.org_id == org_id,
                        SalesforceSchemaTerm.term == term,
                    )
                    .order_by(SalesforceSchemaTerm.weight.desc())
                    .limit(limit)
                )
            ).all()
            if not term_rows:
                continue

            field_refs = [row[0] for row in term_rows if row[0]]
            pairs: List[Tuple[str, str]] = []
            for ref in field_refs:
                if "." in ref:
                    obj, fld = ref.split(".", 1)
                    pairs.append((obj, fld))

            field_meta: Dict[Tuple[str, str], Dict[str, Any]] = {}
            if pairs:
                conditions = [
                    (SalesforceSchemaField.object_api == obj)
                    & (SalesforceSchemaField.field_api == fld)
                    for obj, fld in pairs
                ]
                where_clause = (
                    or_(*conditions) if len(conditions) > 1 else conditions[0]
                )
                field_rows = (
                    await self.db.execute(
                        select(
                            SalesforceSchemaField.object_api,
                            SalesforceSchemaField.field_api,
                            SalesforceSchemaField.label,
                        ).where(
                            SalesforceSchemaField.org_id == org_id,
                            where_clause,
                        )
                    )
                ).all()
                for obj, fld, label in field_rows:
                    field_meta[(obj, fld)] = {
                        "label": label,
                        "object_api": obj,
                        "field_api": fld,
                    }

            if field_refs:
                usage_rows = (
                    await self.db.execute(
                        select(
                            SalesforceSchemaFieldUsage.field_ref,
                            SalesforceSchemaFieldUsage.report_count,
                        ).where(
                            SalesforceSchemaFieldUsage.org_id == org_id,
                            SalesforceSchemaFieldUsage.field_ref.in_(field_refs),
                        )
                    )
                ).all()
                profile_rows = (
                    await self.db.execute(
                        select(
                            SalesforceSchemaFieldProfile.field_ref,
                            SalesforceSchemaFieldProfile.not_null_rate,
                        ).where(
                            SalesforceSchemaFieldProfile.org_id == org_id,
                            SalesforceSchemaFieldProfile.field_ref.in_(field_refs),
                        )
                    )
                ).all()
            else:
                usage_rows = []
                profile_rows = []
            usage_map = {ref: count for ref, count in usage_rows}
            profile_map = {ref: rate for ref, rate in profile_rows}

            for field_ref, weight in term_rows:
                if field_ref not in seen or weight > seen[field_ref]:
                    obj, fld = (
                        field_ref.split(".", 1) if "." in field_ref else ("", field_ref)
                    )
                    meta = field_meta.get(
                        (obj, fld), {"label": fld, "object_api": obj, "field_api": fld}
                    )
                    seen[field_ref] = weight
                    results.append(
                        (
                            field_ref,
                            float(weight),
                            meta.get("label") or meta.get("field_api"),
                            meta.get("object_api"),
                            meta.get("field_api"),
                            int(usage_map.get(field_ref, 0) or 0),
                            float(profile_map.get(field_ref, 0.0) or 0.0),
                        )
                    )
        # Deduplicate while preserving the highest weight
        unique = {}
        for item in results:
            if item[0] not in unique or item[1] > unique[item[0]][1]:
                unique[item[0]] = item
        ranked = sorted(unique.values(), key=lambda row: row[1], reverse=True)
        return ranked[:limit]

    async def mark_full_indexed(self, org_id: str, schema_hash: Optional[str]) -> None:
        await self.upsert_state(
            org_id,
            schema_hash=schema_hash,
            status="ready",
            full_indexed_at=self._utcnow_naive(),
        )
        await self.clear_in_progress(org_id, token=None)
        await self.set_checkpoint(org_id, {})

    async def status(self, org_id: str, *, ttl_hours: int = 24) -> dict:
        obj_count = (
            await self.db.execute(
                select(func.count())
                .select_from(SalesforceSchemaObject)
                .where(SalesforceSchemaObject.org_id == org_id)
            )
        ).scalar_one()
        field_count = (
            await self.db.execute(
                select(func.count())
                .select_from(SalesforceSchemaField)
                .where(SalesforceSchemaField.org_id == org_id)
            )
        ).scalar_one()
        state = await self.get_state(org_id)
        now = self._utcnow_naive()
        ttl = timedelta(hours=max(1, ttl_hours))
        full_dt = self._as_naive(state.full_indexed_at) if state else None
        quick_dt = self._as_naive(state.quick_indexed_at) if state else None
        if state and state.in_progress:
            index_status = "building"
        elif state and full_dt and (now - full_dt) <= ttl:
            index_status = "full"
        elif state and quick_dt and (now - quick_dt) <= ttl:
            index_status = "quick"
        else:
            index_status = "stale"
        source = "index" if index_status in {"full", "quick"} else "live"
        return {
            "org_id": org_id,
            "objects": obj_count,
            "fields": field_count,
            "schema_hash": state.schema_hash if state else None,
            "quick_indexed_at": state.quick_indexed_at.isoformat()
            if state and state.quick_indexed_at
            else None,
            "full_indexed_at": state.full_indexed_at.isoformat()
            if state and state.full_indexed_at
            else None,
            "status": state.status if state else "unknown",
            "in_progress": bool(state.in_progress) if state else False,
            "stage": state.stage if state else None,
            "index_status": index_status,
            "index_source": source,
            "fls_hash": state.fls_hash if state else None,
        }

    async def field_usage_score(self, org_id: str, field_ref: str) -> float:
        """Compute a basic usage score across report + app usage."""
        report_stmt = select(SalesforceSchemaFieldUsage.report_count).where(
            SalesforceSchemaFieldUsage.org_id == org_id,
            SalesforceSchemaFieldUsage.field_ref == field_ref,
        )
        report_count = (await self.db.execute(report_stmt)).scalar_one_or_none() or 0

        app_stmt = select(
            SalesforceSchemaFieldAppUsage.query_count_30d
            + SalesforceSchemaFieldAppUsage.preview_count_30d
        ).where(
            SalesforceSchemaFieldAppUsage.org_id == org_id,
            SalesforceSchemaFieldAppUsage.field_ref == field_ref,
        )
        app_count = (await self.db.execute(app_stmt)).scalar_one_or_none() or 0
        return float(report_count) * 0.7 + float(app_count) * 0.3

    async def field_profile(
        self, org_id: str, field_ref: str
    ) -> Optional[SalesforceSchemaFieldProfile]:
        stmt = select(SalesforceSchemaFieldProfile).where(
            SalesforceSchemaFieldProfile.org_id == org_id,
            SalesforceSchemaFieldProfile.field_ref == field_ref,
        )
        return (await self.db.execute(stmt)).scalar_one_or_none()

    async def search_fields(
        self, org_id: str, term: str, limit: int = 10
    ) -> List[SalesforceSchemaField]:
        """Search by term fallback (like contains)."""
        like_term = f"%{term.lower()}%"
        stmt = (
            select(SalesforceSchemaField)
            .where(
                SalesforceSchemaField.org_id == org_id,
                func.lower(SalesforceSchemaField.label).like(like_term)
                | func.lower(SalesforceSchemaField.field_api).like(like_term),
            )
            .limit(limit)
        )
        return list((await self.db.execute(stmt)).scalars().all())

    async def list_relationships(
        self, org_id: str, from_object: str
    ) -> List[SalesforceSchemaRelationship]:
        stmt = (
            select(SalesforceSchemaRelationship)
            .where(
                SalesforceSchemaRelationship.org_id == org_id,
                SalesforceSchemaRelationship.from_object == from_object,
            )
            .order_by(SalesforceSchemaRelationship.relationship_api.asc())
        )
        return list((await self.db.execute(stmt)).scalars().all())



"""Sequence collision scout check."""

from __future__ import annotations

import time
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any, Dict, List, Tuple

from .base import BaseScoutCheck, PreflightResult
from ..crm_provider import Filter, FilterExpr, FilterOp
from ..models import EvidenceRow, RecommendedAction, ScoutFinding

if TYPE_CHECKING:
    from g_gremlin.outreach import OutreachClient


class SequenceCollisionCheck(BaseScoutCheck):
    _OUTREACH_ACTIVE_STATES = {"active", "in_progress", "pending", "scheduled", "running"}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._current_object = "Lead"
        self._outreach_client: OutreachClient | None = None
        self._outreach_warning_emitted = False
        self._sequence_lookup_cache: dict[str, dict[str, Any]] = {}
        self._recommended_actions: list[RecommendedAction] = []
        self._action_keys: set[tuple[str, str, str]] = set()

    def preflight(self) -> PreflightResult:
        base = super().preflight()
        if base.unsupported:
            return base

        target_objects = ["Lead"]
        if bool(self.params.get("include_contacts", True)):
            target_objects.append("Contact")

        for object_type in target_objects:
            if not self.provider.supports_object(object_type):
                self._warnings.append(f"Unsupported object: {object_type}")
                continue
            available = self._available_fields(object_type)
            if "email" not in available:
                self._warnings.append(f"{object_type}: Email field unavailable; collision scan skipped for this object.")
        return PreflightResult(ok=True)

    def build_count_filters(self) -> Tuple[str, List[FilterExpr]]:
        object_type = self._current_object
        return object_type, self._base_filters_for_object(object_type, self._available_fields(object_type))

    def build_sample_request(self) -> Tuple[str, List[str], List[FilterExpr], str | None, str, bool]:
        object_type = self._current_object
        available = self._available_fields(object_type)
        fields, order_by = self._fields_for_object(object_type, available)
        filters = self._base_filters_for_object(object_type, available)
        return object_type, fields, filters, order_by, "DESC", False

    def evaluate(self, rows: List[Dict[str, Any]]) -> List[EvidenceRow]:
        evidence: List[EvidenceRow] = []
        for row in rows:
            candidate = self._evaluate_row(row)
            if candidate is not None:
                evidence.append(candidate)
        return evidence

    def execute(self) -> ScoutFinding:  # type: ignore[override]
        start = time.perf_counter()
        self._recommended_actions = []
        self._action_keys = set()
        self._sequence_lookup_cache = {}

        try:
            preflight = self.preflight()
            if preflight.unsupported:
                return self._unsupported_finding(preflight.reason or "Unsupported")

            target_objects = ["Lead"]
            if bool(self.params.get("include_contacts", True)):
                target_objects.append("Contact")

            self._outreach_client = self._build_outreach_client()

            total_scanned = 0
            full_evidence: list[EvidenceRow] = []
            query_limit = self._query_limit()

            for object_type in target_objects:
                if not self.provider.supports_object(object_type):
                    continue

                self._current_object = object_type
                available = self._available_fields(object_type)
                if "email" not in available:
                    continue

                filters = self._base_filters_for_object(object_type, available)
                count = self.provider.count(object_type, filters)
                self._collect_provider_meta()
                total_scanned += count
                if count <= 0:
                    continue

                fields, order_by = self._fields_for_object(object_type, available)
                if count > query_limit:
                    self._warnings.append(
                        f"{object_type}: scan capped at {query_limit} rows out of {count}; findings may be partial."
                    )

                rows = self.provider.query(
                    object_type,
                    fields,
                    filters,
                    order_by=order_by,
                    order_dir="DESC",
                    nulls_first=False,
                    limit=query_limit,
                )
                self._collect_provider_meta()
                typed_rows = [self._attach_sobject(row, object_type) for row in rows]

                if self._outreach_client is None:
                    continue
                full_evidence.extend(self.evaluate(typed_rows))

            findings_count = len(full_evidence)
            evidence = full_evidence[: self.top_n]
            status = "findings" if findings_count > 0 else "ok"
            duration_ms = int((time.perf_counter() - start) * 1000)

            return ScoutFinding(
                scout_name=self.definition.name,
                display_name=self.definition.display_name,
                description=self.definition.description,
                category=self.definition.category,
                severity=self.definition.severity_default,
                total_records_scanned=total_scanned,
                findings_count=findings_count,
                evidence=evidence,
                recommended_actions=self.get_recommended_actions(),
                soql_queries=self._soql_queries,
                duration_ms=duration_ms,
                parameters_used=self.params,
                status=status,
                warnings=self._warnings,
            )
        except Exception as exc:
            duration_ms = int((time.perf_counter() - start) * 1000)
            return self._error_finding(str(exc), duration_ms=duration_ms)

    def get_recommended_actions(self) -> List[RecommendedAction]:
        return list(self._recommended_actions)

    def _evaluate_row(self, row: Dict[str, Any]) -> EvidenceRow | None:
        record_id = str(row.get("Id") or "").strip()
        if not record_id:
            return None

        email = str(row.get("Email") or "").strip().lower()
        if not email:
            return None

        context = self._lookup_sequence_context(email)
        active_states = list(context.get("active_sequence_states", []))
        min_active = max(2, self._to_int(self.params.get("min_active_sequences"), default=2))
        if len(active_states) < min_active:
            return None

        keeper, pause_candidates, priority_reason = self._select_priority(active_states)
        if keeper is None or not pause_candidates:
            return None

        active_state_ids = [str(item.get("sequence_state_id") or "") for item in active_states]
        active_sequence_ids = [str(item.get("sequence_id") or "") for item in active_states if item.get("sequence_id")]
        pause_state_ids = [str(item.get("sequence_state_id") or "") for item in pause_candidates]
        pause_sequence_ids = [str(item.get("sequence_id") or "") for item in pause_candidates if item.get("sequence_id")]

        sobject = str(row.get("_sobject") or self._current_object or "Lead")
        reason = (
            f"Prospect is enrolled in {len(active_states)} active sequences; "
            f"keep one enrollment and pause the lower-priority overlaps."
        )
        fields_payload: dict[str, Any] = {
            "Name": row.get("Name"),
            "Email": email,
            "OwnerId": row.get("OwnerId"),
            "active_sequence_count": len(active_states),
            "active_sequence_ids": active_sequence_ids,
            "active_sequence_state_ids": active_state_ids,
            "keeper_sequence_state_id": keeper.get("sequence_state_id"),
            "keeper_sequence_id": keeper.get("sequence_id"),
            "keeper_sequence_name": keeper.get("sequence_name"),
            "pause_sequence_state_ids": pause_state_ids,
            "pause_sequence_ids": pause_sequence_ids,
            "priority_reason": priority_reason,
        }

        prospect_id = context.get("prospect_id")
        related_ids = self._dedupe_non_blank(active_state_ids + active_sequence_ids + [str(prospect_id or "")])
        if prospect_id:
            fields_payload["outreach_prospect_id"] = str(prospect_id)

        self._add_action(
            record_id=record_id,
            sobject=sobject,
            prospect_id=str(prospect_id or ""),
            email=email,
            keeper=keeper,
            pause_candidates=pause_candidates,
            priority_reason=priority_reason,
        )

        return EvidenceRow(
            record_id=record_id,
            sobject=sobject,
            fields=fields_payload,
            reason=reason,
            reason_code="SEQUENCE_COLLISION",
            severity=self.definition.severity_default,
            related_ids=related_ids,
        )

    def _add_action(
        self,
        *,
        record_id: str,
        sobject: str,
        prospect_id: str,
        email: str,
        keeper: dict[str, Any],
        pause_candidates: list[dict[str, Any]],
        priority_reason: str,
    ) -> None:
        keeper_state_id = str(keeper.get("sequence_state_id") or "")
        pause_state_ids = [str(item.get("sequence_state_id") or "") for item in pause_candidates]
        pause_state_ids = [value for value in pause_state_ids if value]
        key_anchor = prospect_id or email
        action_key = (key_anchor, keeper_state_id, ",".join(sorted(pause_state_ids)))
        if not key_anchor or not keeper_state_id or not pause_state_ids or action_key in self._action_keys:
            return

        self._action_keys.add(action_key)
        pause_sequence_ids = [str(item.get("sequence_id") or "") for item in pause_candidates if item.get("sequence_id")]
        self._recommended_actions.append(
            RecommendedAction(
                action_type="pause_sequence",
                label="Pause lower-priority sequence enrollment",
                risk_level="low",
                details={
                    "reason_code": "SEQUENCE_COLLISION",
                    "record_id": record_id,
                    "sobject": sobject,
                    "prospect_id": key_anchor,
                    "keep_sequence_state_id": keeper_state_id,
                    "keep_sequence_id": keeper.get("sequence_id"),
                    "pause_sequence_state_ids": pause_state_ids,
                    "pause_sequence_ids": pause_sequence_ids,
                    "priority_reason": priority_reason,
                },
            )
        )

    def _build_outreach_client(self) -> OutreachClient | None:
        try:
            from g_gremlin.outreach import OutreachClient

            return OutreachClient()
        except Exception:
            if not self._outreach_warning_emitted:
                self._warnings.append(
                    "Outreach not configured; sequence collision scan runs in no-op mode."
                )
                self._outreach_warning_emitted = True
            return None

    def _lookup_sequence_context(self, email: str) -> dict[str, Any]:
        blank = {
            "outreach_connected": bool(self._outreach_client),
            "prospect_id": None,
            "active_sequence_states": [],
        }
        normalized = str(email or "").strip().lower()
        if not normalized:
            return dict(blank)
        if normalized in self._sequence_lookup_cache:
            return dict(self._sequence_lookup_cache[normalized])
        if self._outreach_client is None:
            self._sequence_lookup_cache[normalized] = dict(blank)
            return dict(blank)

        try:
            prospect = self._outreach_client.find_prospect_by_email(normalized)
            if prospect is None or not getattr(prospect, "id", None):
                resolved = dict(blank)
                resolved["outreach_connected"] = True
                self._sequence_lookup_cache[normalized] = resolved
                return dict(resolved)

            prospect_id = str(prospect.id)
            states = self._outreach_client.get_sequence_states_for_prospect(prospect_id)
            active_states: list[dict[str, Any]] = []
            for item in states:
                parsed = self._parse_active_sequence_state(item)
                if parsed is not None:
                    active_states.append(parsed)

            resolved = {
                "outreach_connected": True,
                "prospect_id": prospect_id,
                "active_sequence_states": active_states,
            }
            self._sequence_lookup_cache[normalized] = resolved
            return dict(resolved)
        except Exception as exc:
            self._warnings.append(f"Outreach sequence lookup failed for {normalized}: {exc}")
            unresolved = dict(blank)
            unresolved["outreach_connected"] = False
            self._sequence_lookup_cache[normalized] = unresolved
            return dict(unresolved)

    def _parse_active_sequence_state(self, state: Any) -> dict[str, Any] | None:
        if not isinstance(state, dict):
            return None
        attributes = state.get("attributes") if isinstance(state.get("attributes"), dict) else {}
        state_value = str(attributes.get("state") or "").strip().lower()
        if state_value not in self._OUTREACH_ACTIVE_STATES:
            return None

        state_id = str(state.get("id") or "").strip()
        if not state_id:
            return None

        sequence_id = self._sequence_id_from_state(state)
        created_at = self._extract_state_timestamp(attributes)
        return {
            "sequence_state_id": state_id,
            "sequence_id": sequence_id,
            "sequence_name": self._sequence_name_from_state(state),
            "state": state_value,
            "created_at": created_at,
        }

    def _sequence_id_from_state(self, state: dict[str, Any]) -> str:
        relationships = state.get("relationships") if isinstance(state.get("relationships"), dict) else {}
        sequence_rel = relationships.get("sequence") if isinstance(relationships, dict) else {}
        sequence_data = sequence_rel.get("data") if isinstance(sequence_rel, dict) else {}
        sequence_id = sequence_data.get("id") if isinstance(sequence_data, dict) else None
        if sequence_id:
            return str(sequence_id)

        attributes = state.get("attributes") if isinstance(state.get("attributes"), dict) else {}
        for key in ("sequenceId", "sequence_id"):
            value = attributes.get(key)
            if value:
                return str(value)
        return ""

    def _sequence_name_from_state(self, state: dict[str, Any]) -> str:
        attributes = state.get("attributes") if isinstance(state.get("attributes"), dict) else {}
        for key in ("sequenceName", "sequence_name", "name"):
            value = attributes.get(key)
            if value:
                return str(value)
        return ""

    def _extract_state_timestamp(self, attributes: dict[str, Any]) -> str:
        for key in ("createdAt", "created_at", "createdDate", "created"):
            value = attributes.get(key)
            if value:
                return str(value)
        return ""

    def _select_priority(
        self, active_states: list[dict[str, Any]]
    ) -> tuple[dict[str, Any] | None, list[dict[str, Any]], str]:
        def sort_key(item: dict[str, Any]) -> tuple[float, str, str]:
            created_at = self._parse_iso_datetime(item.get("created_at"))
            created_rank = created_at.timestamp() if created_at is not None else float("inf")
            sequence_id = str(item.get("sequence_id") or "")
            state_id = str(item.get("sequence_state_id") or "")
            return (created_rank, sequence_id, state_id)

        ordered = sorted(active_states, key=sort_key)
        if not ordered:
            return None, [], "stable_id_order"
        keeper = ordered[0]
        reason = "oldest_active_enrollment"
        if self._parse_iso_datetime(keeper.get("created_at")) is None:
            reason = "stable_id_order"
        return keeper, ordered[1:], reason

    def _base_filters_for_object(
        self, object_type: str, available_fields: set[str] | None = None
    ) -> list[FilterExpr]:
        available = available_fields if available_fields is not None else self._available_fields(object_type)
        filters: list[FilterExpr] = [Filter("Email", FilterOp.IS_NOT_NULL)]
        lookback_days = max(0, self._to_int(self.params.get("lookback_days"), default=30))
        if lookback_days <= 0:
            return filters
        if "lastmodifieddate" in available:
            filters.append(Filter("LastModifiedDate", FilterOp.LAST_N_DAYS, lookback_days))
        elif "createddate" in available:
            filters.append(Filter("CreatedDate", FilterOp.LAST_N_DAYS, lookback_days))
        return filters

    def _fields_for_object(self, object_type: str, available_fields: set[str]) -> tuple[list[str], str | None]:
        fields: list[str] = ["Id"]
        for candidate in ("Name", "Email", "OwnerId", "AccountId", "Company"):
            if candidate.lower() in available_fields and candidate not in fields:
                fields.append(candidate)

        order_field = None
        if "lastmodifieddate" in available_fields:
            order_field = "LastModifiedDate"
        elif "createddate" in available_fields:
            order_field = "CreatedDate"
        if order_field and order_field not in fields:
            fields.append(order_field)
        return fields, order_field

    def _query_limit(self) -> int:
        default_limit = max(self.top_n * 5, 100)
        configured = self._to_int(self.params.get("max_candidates"), default=default_limit)
        return max(self.top_n, configured)

    def _available_fields(self, object_type: str) -> set[str]:
        available_method = getattr(self.provider, "available_fields", None)
        if callable(available_method):
            try:
                return {str(name).lower() for name in available_method(object_type)}
            except Exception:
                pass

        try:
            describe = self.provider.describe(object_type)
        except Exception:
            return set()
        fields = describe.get("result", {}).get("fields", []) if isinstance(describe, dict) else []
        if not isinstance(fields, list):
            return set()
        return {
            str(item.get("name")).lower()
            for item in fields
            if isinstance(item, dict) and item.get("name")
        }

    def _parse_iso_datetime(self, value: Any) -> datetime | None:
        text = str(value or "").strip()
        if not text:
            return None
        try:
            parsed = datetime.fromisoformat(text.replace("Z", "+00:00"))
        except Exception:
            return None
        if parsed.tzinfo is None:
            return parsed.replace(tzinfo=timezone.utc)
        return parsed.astimezone(timezone.utc)

    def _to_int(self, value: Any, *, default: int) -> int:
        try:
            return int(value)
        except Exception:
            return default

    def _attach_sobject(self, row: Dict[str, Any], object_type: str) -> Dict[str, Any]:
        copy = dict(row)
        copy["_sobject"] = object_type
        return copy

    def _dedupe_non_blank(self, values: list[str]) -> list[str]:
        ordered: list[str] = []
        seen: set[str] = set()
        for value in values:
            token = str(value or "").strip()
            if not token or token in seen:
                continue
            seen.add(token)
            ordered.append(token)
        return ordered


__all__ = ["SequenceCollisionCheck"]



"""
KMS Envelope Encryption Service for BYO Provider Locker.

Implements envelope encryption pattern:
1. Generate random 256-bit DEK (data encryption key)
2. Encrypt secret with DEK using AES-256-GCM
3. Encrypt DEK with Cloud KMS (store encrypted_dek + kms_key_version)
4. To decrypt: KMS decrypt DEK -> AES-GCM decrypt secret
5. Rotation: rewrap DEKs with new KMS version without touching ciphertext
"""

from __future__ import annotations

import os
import secrets
from dataclasses import dataclass
from typing import Optional

from cryptography.hazmat.primitives.ciphers.aead import AESGCM
import logging

logger = logging.getLogger(__name__)

# Try to import Google Cloud KMS - graceful fallback for local dev
try:
    from google.cloud import kms
    HAS_KMS = True
except ImportError:
    HAS_KMS = False
    kms = None

# Constants
DEK_SIZE_BYTES = 32  # 256-bit key for AES-256
NONCE_SIZE_BYTES = 12  # 96-bit nonce for GCM (recommended)


@dataclass
class EncryptedSecret:
    """Container for envelope-encrypted secret data."""
    encrypted_value: bytes  # AES-256-GCM encrypted secret
    encryption_nonce: bytes  # 12-byte GCM nonce
    encrypted_dek: bytes  # DEK wrapped by KMS
    kms_key_version: str  # KMS key version for rotation tracking


@dataclass
class KMSConfig:
    """Configuration for Google Cloud KMS."""
    project_id: str
    location: str
    key_ring: str
    crypto_key: str

    @property
    def key_name(self) -> str:
        """Full resource name for the KMS key."""
        return (
            f"projects/{self.project_id}/locations/{self.location}/"
            f"keyRings/{self.key_ring}/cryptoKeys/{self.crypto_key}"
        )

    @classmethod
    def from_env(cls) -> "KMSConfig":
        """Load KMS config from environment variables."""
        return cls(
            project_id=os.getenv("KMS_PROJECT_ID") or os.getenv("GOOGLE_CLOUD_PROJECT", ""),
            location=os.getenv("KMS_LOCATION", "us-central1"),
            key_ring=os.getenv("KMS_KEY_RING", "foundrymatch-secrets"),
            crypto_key=os.getenv("KMS_CRYPTO_KEY", "workspace-secrets"),
        )


class KMSEnvelopeEncryption:
    """
    Envelope encryption service using Google Cloud KMS.

    For local development without KMS, uses a fallback mode that stores
    DEKs directly (encrypted with local key). This is NOT secure for production.
    """

    def __init__(
        self,
        config: Optional[KMSConfig] = None,
        allow_local_fallback: bool = False,
    ):
        """
        Initialize KMS envelope encryption.

        Args:
            config: KMS configuration. If None, loads from environment.
            allow_local_fallback: If True, use local encryption when KMS unavailable.
                                  Only enable for development/testing.
        """
        self.config = config or KMSConfig.from_env()
        self.allow_local_fallback = allow_local_fallback
        self._kms_client: Optional["kms.KeyManagementServiceClient"] = None

        # Local fallback key (only for development)
        self._local_fallback_key: Optional[bytes] = None
        if allow_local_fallback:
            fallback_key = os.getenv("ENCRYPTION_KEY", "")
            if fallback_key:
                # Derive 32-byte key from whatever key is provided
                from hashlib import sha256
                self._local_fallback_key = sha256(fallback_key.encode()).digest()

    @property
    def kms_client(self) -> "kms.KeyManagementServiceClient":
        """Lazy-load KMS client."""
        if self._kms_client is None:
            if not HAS_KMS:
                raise RuntimeError(
                    "google-cloud-kms is not installed. "
                    "Install with: pip install google-cloud-kms"
                )
            self._kms_client = kms.KeyManagementServiceClient()
        return self._kms_client

    def _use_local_fallback(self) -> bool:
        """Check if we should use local fallback mode."""
        if not self.allow_local_fallback:
            return False
        if not self._local_fallback_key:
            return False
        # Use fallback if KMS isn't available or configured
        if not HAS_KMS:
            return True
        if not self.config.project_id:
            return True
        return False

    def _local_wrap_dek(self, dek: bytes) -> tuple[bytes, str]:
        """
        Wrap DEK using local fallback (development only).

        WARNING: This is NOT secure for production. DEK is encrypted with
        a local key derived from ENCRYPTION_KEY, not KMS.
        """
        if not self._local_fallback_key:
            raise RuntimeError("Local fallback key not configured")

        nonce = secrets.token_bytes(NONCE_SIZE_BYTES)
        aesgcm = AESGCM(self._local_fallback_key)
        encrypted_dek = nonce + aesgcm.encrypt(nonce, dek, None)
        return encrypted_dek, "local-fallback-v1"

    def _local_unwrap_dek(self, encrypted_dek: bytes, key_version: str) -> bytes:
        """
        Unwrap DEK using local fallback (development only).
        """
        if not self._local_fallback_key:
            raise RuntimeError("Local fallback key not configured")
        if key_version != "local-fallback-v1":
            raise ValueError(f"Unknown local key version: {key_version}")

        nonce = encrypted_dek[:NONCE_SIZE_BYTES]
        ciphertext = encrypted_dek[NONCE_SIZE_BYTES:]
        aesgcm = AESGCM(self._local_fallback_key)
        return aesgcm.decrypt(nonce, ciphertext, None)

    def _kms_wrap_dek(self, dek: bytes) -> tuple[bytes, str]:
        """
        Wrap DEK using Google Cloud KMS.

        Returns:
            Tuple of (encrypted_dek, kms_key_version)
        """
        response = self.kms_client.encrypt(
            request={
                "name": self.config.key_name,
                "plaintext": dek,
            }
        )
        # Extract version from the response name
        # Format: projects/.../cryptoKeys/.../cryptoKeyVersions/1
        version = response.name.rsplit("/", 1)[-1] if "/" in response.name else "1"
        return response.ciphertext, f"kms-v{version}"

    def _kms_unwrap_dek(self, encrypted_dek: bytes, key_version: str) -> bytes:
        """
        Unwrap DEK using Google Cloud KMS.
        """
        # Cloud KMS symmetric decrypt expects the CryptoKey resource name, not
        # CryptoKeyVersion. The ciphertext carries version metadata internally.
        # We still persist key_version for audit/rotation reporting.
        response = self.kms_client.decrypt(
            request={
                "name": self.config.key_name,
                "ciphertext": encrypted_dek,
            }
        )
        return response.plaintext

    def encrypt(self, plaintext: str) -> EncryptedSecret:
        """
        Encrypt a secret using envelope encryption.

        Args:
            plaintext: The secret to encrypt (e.g., API key)

        Returns:
            EncryptedSecret containing all components needed for storage
        """
        # Step 1: Generate random DEK
        dek = secrets.token_bytes(DEK_SIZE_BYTES)

        # Step 2: Encrypt plaintext with DEK using AES-256-GCM
        nonce = secrets.token_bytes(NONCE_SIZE_BYTES)
        aesgcm = AESGCM(dek)
        encrypted_value = aesgcm.encrypt(nonce, plaintext.encode("utf-8"), None)

        # Step 3: Wrap DEK with KMS (or local fallback)
        if self._use_local_fallback():
            logger.warning(
                "Using local fallback encryption - NOT SECURE FOR PRODUCTION"
            )
            encrypted_dek, kms_key_version = self._local_wrap_dek(dek)
        else:
            encrypted_dek, kms_key_version = self._kms_wrap_dek(dek)

        # Step 4: Clear DEK from memory
        # Note: Python doesn't guarantee immediate clearing, but we try
        dek = b"\x00" * DEK_SIZE_BYTES  # noqa: F841

        return EncryptedSecret(
            encrypted_value=encrypted_value,
            encryption_nonce=nonce,
            encrypted_dek=encrypted_dek,
            kms_key_version=kms_key_version,
        )

    def decrypt(self, secret: EncryptedSecret) -> str:
        """
        Decrypt an envelope-encrypted secret.

        Args:
            secret: EncryptedSecret from storage

        Returns:
            Decrypted plaintext string
        """
        # Step 1: Unwrap DEK
        if secret.kms_key_version.startswith("local-fallback"):
            dek = self._local_unwrap_dek(secret.encrypted_dek, secret.kms_key_version)
        else:
            dek = self._kms_unwrap_dek(secret.encrypted_dek, secret.kms_key_version)

        # Step 2: Decrypt secret with DEK
        aesgcm = AESGCM(dek)
        plaintext_bytes = aesgcm.decrypt(
            secret.encryption_nonce,
            secret.encrypted_value,
            None,
        )

        # Step 3: Clear DEK from memory
        dek = b"\x00" * DEK_SIZE_BYTES  # noqa: F841

        return plaintext_bytes.decode("utf-8")

    def rewrap_dek(self, secret: EncryptedSecret) -> EncryptedSecret:
        """
        Re-wrap DEK with current KMS key version (for key rotation).

        This allows rotating KMS keys without re-encrypting the actual secrets.
        Only the DEK wrapper changes.

        Args:
            secret: Existing EncryptedSecret to rewrap

        Returns:
            New EncryptedSecret with same encrypted_value but new DEK wrapper
        """
        # Unwrap DEK with old key version
        if secret.kms_key_version.startswith("local-fallback"):
            dek = self._local_unwrap_dek(secret.encrypted_dek, secret.kms_key_version)
        else:
            dek = self._kms_unwrap_dek(secret.encrypted_dek, secret.kms_key_version)

        # Wrap with current key version
        if self._use_local_fallback():
            encrypted_dek, kms_key_version = self._local_wrap_dek(dek)
        else:
            encrypted_dek, kms_key_version = self._kms_wrap_dek(dek)

        # Clear DEK
        dek = b"\x00" * DEK_SIZE_BYTES  # noqa: F841

        return EncryptedSecret(
            encrypted_value=secret.encrypted_value,
            encryption_nonce=secret.encryption_nonce,
            encrypted_dek=encrypted_dek,
            kms_key_version=kms_key_version,
        )


# Global instance (lazy initialization)
_envelope_encryption: Optional[KMSEnvelopeEncryption] = None


def get_envelope_encryption(allow_local_fallback: Optional[bool] = None) -> KMSEnvelopeEncryption:
    """
    Get or create the global envelope encryption instance.

    Args:
        allow_local_fallback: Override for local fallback setting.
                              If None, uses ENV != "production" as default.
    """
    global _envelope_encryption
    if _envelope_encryption is None:
        if allow_local_fallback is None:
            # Default: allow fallback in non-production environments
            allow_local_fallback = os.getenv("ENV", "development") != "production"
        _envelope_encryption = KMSEnvelopeEncryption(
            allow_local_fallback=allow_local_fallback
        )
    return _envelope_encryption


def encrypt_secret(plaintext: str) -> EncryptedSecret:
    """Convenience function to encrypt a secret."""
    return get_envelope_encryption().encrypt(plaintext)


def decrypt_secret(secret: EncryptedSecret) -> str:
    """Convenience function to decrypt a secret."""
    return get_envelope_encryption().decrypt(secret)


def get_key_suffix(plaintext: str, length: int = 4) -> str:
    """
    Get the last N characters of a secret for display purposes.

    Args:
        plaintext: The secret value
        length: Number of characters to show (default 4)

    Returns:
        Last N characters of the secret
    """
    if len(plaintext) <= length:
        return "*" * len(plaintext)
    return plaintext[-length:]
